/*
**  Copyright (c) 2003,2004,2005 Andrea Luzzardi <scox@gentoo.org>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "pam.h"
#include "parser.h"
#include "conf.h"
#include "device.h"
#include "dsa.h"

int pam_conv_pass (pam_handle_t *pamh, const char *prompt)
{
	const struct pam_conv *conv;
	struct pam_message msg;
	const struct pam_message *msgs[1];
	struct pam_response *resp;
	const void *item;

	if (pam_get_item (pamh, PAM_CONV, &item) != PAM_SUCCESS)
		return 0;
	
	conv = (const struct pam_conv *)item;

	msg.msg_style = PAM_PROMPT_ECHO_OFF;
	msg.msg = prompt;
	msgs[0] = &msg;

	if ((conv->conv (1, msgs, &resp, conv->appdata_ptr)) != PAM_SUCCESS)
		return 0;

	if ((pam_set_item (pamh, PAM_AUTHTOK, resp[0].resp)) != PAM_SUCCESS)
		return 0;
	
	memset (resp[0].resp, 0, strlen (resp[0].resp));
	free (resp[0].resp);
	free (resp);

	return 1;	
}

int pam_get_pass (pam_handle_t *pamh, const char **passp, const char *prompt)
{
	const void *item = NULL;

	if ((usb_conf.try_first_pass == 1) || (usb_conf.use_first_pass == 1)) {
		if (pam_get_item (pamh, PAM_AUTHTOK, &item) != PAM_SUCCESS)
			return 0;
	}

	if (!item) {
		if (usb_conf.use_first_pass == 1) {
			DEBUGP ("Cannot import the passphrase\n");
			return 0;
		}

		if (!pam_conv_pass (pamh, prompt))
			return 0;

		if (pam_get_item (pamh, PAM_AUTHTOK, &item) != PAM_SUCCESS)
			return 0;
	} else {
		DEBUGP ("Importing the passphrase...\n");
	}
	
	*passp = (const char *)item;

	return 1;
}

int check_serial (FILE *proc)
{
	FILE *f;
	char *line = NULL;
	char *serial = NULL;
	char type[] = "{S}";

	/* If there's no serial number file, authorize */
	if ((f = fopen (usb_conf.snfile, "r")) == NULL)
		return 1;

	DEBUGP ("Found serial number ACL file, checking...\n");

	serial = get_param (proc, "Serial Number", ':');

	if (!serial || !(strcmp (serial, "None"))) {
		free (serial);
		serial = NULL;
		/* trying with the GUID */
		DEBUGP ("Cannot fetch the device's serial number\n"
			"Checking the GUID...\n");
		type[1] = 'G';
		if (!(serial = get_param (proc, "GUID", ':'))) {
			/* no luck either */
			free (serial);
			fclose (f);
			return 0;
		}
		DEBUGP ("GUID found\n");
	}

	while ((line = get_param (f, type, ' '))) {
		
		if (!strcmp (serial, line)) {
			free (serial);
			free (line);
			fclose (f);
			DEBUGP ("Serial number matches\n");
			return 1;
		}

		free (line);
	}

	free (serial);
	fclose (f);
	DEBUGP ("Serial number doesn't match\n");
	return 0;
}

int check_device (void)
{
	FILE *f;
	char *file;

	if (usb_conf.check_device == -1)
		return 1;

	while ((file = find_proc_file())) {
		DEBUGP("Using procfile %s\n", file);

		if (!(f = fopen (file, "r"))) {
			DEBUGP("Cannot open procfile %s\n", file);
			free (file);
			return 0;
		}

		if (!check_serial (f)) {
			DEBUGP("Invalid serial number\n");
			fclose (f);
			free (file);
			continue;
		}

		if (!check_param (f, "Attached", "Yes", ':')) {
			DEBUGP("Device found on %s is not attached\n", file);
			fclose (f);
			free (file);
			continue;
		}

		DEBUGP ("Found valid device %s\n", file);
		fclose (f);
		free (file);
		return 1;
	}
	return 0;
}

int authenticate (pam_handle_t* pamh, const char *user)
{
	FILE *pubfd;
	FILE *privfd;

	char *file;

	int status;

	DSA *private;
	DSA *public;

	if (!(privfd = open_dev (user))) {
		INFOP ("Authentication denied: Unable to find a valid device.\n");
		return 0;
	}

	file = get_public_filename (user);

	if (!(pubfd = fopen (file, "r"))) {
		INFOP ("Authentication denied: Cannot open public key \"%s\"\n", file);
		free (file);
		close_dev (privfd);
		return 0;
	}

	if (!(private = import_private_key (privfd, pamh))) {
		INFOP("Authentication denied: Cannot import private key\n");
		free (file);
		close_dev (privfd);
		fclose (pubfd);
		return 0;
	}

	DEBUGP ("Private key imported\n");

	if (!(public = import_public_key (pubfd))) {
		INFOP ("Authentication denied: Cannot import public key \"%s\"\n",
				file);
		free (file);
		close_dev (privfd);
		fclose (pubfd);
		DSA_free (private);
		return 0;
	}

	DEBUGP ("Public key imported\n");

	free (file);
	close_dev (privfd);
	fclose (pubfd);

	status = valid_dsa_keys (private, public);
	
	DSA_free (private);
	DSA_free (public);

	return status;
}
